//
//  CustomDNSViewController.swift
//  MullvadVPN
//
//  Created by Jon Petersson on 2023-11-09.
//  Copyright © 2023 Mullvad VPN AB. All rights reserved.
//

import MullvadSettings
import UIKit

class CustomDNSViewController: UITableViewController {
    private let interactor: VPNSettingsInteractor
    private var dataSource: CustomDNSDataSource?
    private let alertPresenter: AlertPresenter

    override var preferredStatusBarStyle: UIStatusBarStyle {
        .lightContent
    }

    init(interactor: VPNSettingsInteractor, alertPresenter: AlertPresenter) {
        self.interactor = interactor
        self.alertPresenter = alertPresenter

        super.init(style: .grouped)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        tableView.accessibilityIdentifier = .dnsSettingsTableView
        tableView.backgroundColor = .secondaryColor
        tableView.separatorColor = .secondaryColor
        tableView.rowHeight = UITableView.automaticDimension
        tableView.estimatedRowHeight = 60
        tableView.estimatedSectionHeaderHeight = tableView.estimatedRowHeight

        dataSource = CustomDNSDataSource(tableView: tableView)
        dataSource?.delegate = self

        navigationItem.title = NSLocalizedString(
            "NAVIGATION_TITLE",
            tableName: "VPNSettings",
            value: "DNS settings",
            comment: ""
        )

        navigationItem.rightBarButtonItem = editButtonItem
        navigationItem.rightBarButtonItem?.accessibilityIdentifier = .dnsSettingsEditButton

        interactor.tunnelSettingsDidChange = { [weak self] newSettings in
            self?.dataSource?.update(from: newSettings)
        }
        dataSource?.update(from: interactor.tunnelSettings)

        tableView.tableHeaderView = UIView(frame: CGRect(
            origin: .zero,
            size: CGSize(width: 0, height: UIMetrics.TableView.sectionSpacing)
        ))
    }

    override func setEditing(_ editing: Bool, animated: Bool) {
        super.setEditing(editing, animated: animated)

        dataSource?.setEditing(editing, animated: animated)

        navigationItem.setHidesBackButton(editing, animated: animated)

        // Disable swipe to dismiss when editing
        isModalInPresentation = editing
    }

    private func showInfo(with message: NSAttributedString) {
        let presentation = AlertPresentation(
            id: "vpn-settings-content-blockers-alert",
            icon: .info,
            attributedMessage: message,
            buttons: [
                AlertAction(
                    title: NSLocalizedString(
                        "VPN_SETTINGS_DNS_SETTINGS_OK_ACTION",
                        tableName: "ContentBlockers",
                        value: "Got it!",
                        comment: ""
                    ),
                    style: .default
                ),
            ]
        )

        alertPresenter.showAlert(presentation: presentation, animated: true)
    }
}

extension CustomDNSViewController: DNSSettingsDataSourceDelegate {
    func didChangeViewModel(_ viewModel: VPNSettingsViewModel) {
        interactor.updateSettings([.dnsSettings(viewModel.asDNSSettings())])
    }

    func showInfo(for item: VPNSettingsInfoButtonItem) {
        var message = NSAttributedString()

        switch item {
        case .contentBlockers:
            message = NSAttributedString(markdownString: NSLocalizedString(
                "VPN_SETTINGS_CONTENT_BLOCKERS_GENERAL",
                tableName: "ContentBlockers",
                value: """
                When this feature is enabled it stops the device from contacting certain \
                domains or websites known for distributing ads, malware, trackers and more. \

                This might cause issues on certain websites, services, and apps.
                Attention: this setting cannot be used in combination with **Use custom DNS server**.
                """,
                comment: ""
            ), options: MarkdownStylingOptions(font: .preferredFont(forTextStyle: .body)))

        case .blockMalware:
            message = NSAttributedString(markdownString: NSLocalizedString(
                "VPN_SETTINGS_CONTENT_BLOCKERS_MALWARE",
                tableName: "ContentBlockers",
                value: """
                Warning: The malware blocker is not an anti-virus and should not \
                be treated as such, this is just an extra layer of protection.
                """,
                comment: ""
            ), options: MarkdownStylingOptions(font: .preferredFont(forTextStyle: .body)))

        default:
            assertionFailure("No matching InfoButtonItem")
        }

        showInfo(with: message)
    }
}
